#ifndef F_CPU
#define F_CPU 16000000UL
#endif

#include <avr/io.h>
#include <util/delay.h>



/*

Dx_ADDR is the DMX address of the dowser. Range 1-512
Dx_OPEN is the default (not-triggered) position.  Range 0-255
Dx_CLOSED is the triggered position.  Range 0-255

Closed can be less than open.

Digital inputs 2 through 7 are triggered when connected to ground (0V)

D2 => Dowser 1
D3 => Dowser 2
D4 => Dowser 3
D5 => Dowser 4
D6 => Dowser 5
D7 => Dowser 6

*/


#define D1_ADDR 1
#define D1_OPEN 0
#define D1_CLOSED 255

#define D2_ADDR 2
#define D2_OPEN 0
#define D2_CLOSED 255

#define D3_ADDR 3
#define D3_OPEN 0
#define D3_CLOSED 255

#define D4_ADDR 4
#define D4_OPEN 0
#define D4_CLOSED 255

#define D5_ADDR 5
#define D5_OPEN 0
#define D5_CLOSED 255

#define D6_ADDR 6
#define D6_OPEN 0
#define D6_CLOSED 255

/// Don't change things below this line





#ifndef CHANNELS
#define CHANNELS 512
#endif

uint8_t volatile universe[CHANNELS + 1];
uint16_t volatile num_channels;
uint16_t volatile start_address;
uint16_t volatile channel = 0;
uint8_t volatile need_break = 0; 
uint8_t volatile start_packet = 0;
uint8_t volatile ptr = 0;


// Sets address to value, range checks.
void DMXput(uint16_t address, uint8_t value)
{
	if(address >= start_address && address < start_address + num_channels)
	{
		universe[address - start_address] = value;
	}
}

void dmx_break(void) 
{
	// Disable TXD0 and pull low PD1
	PORTD = 0b11111100;
	UCSR0B = 0b00000000;
	_delay_us(20);
        _delay_us(20);
        _delay_us(20);
        _delay_us(20);
                _delay_us(20);
        _delay_us(20);

        _delay_us(20); // DMX-512 defines minimum break as 92us, we need to add 20us for the last slot to finish transmitting after we disable the TX.
}

void dmx_mab(void)
{
	PORTD = 0b11111110; // pull high
	_delay_us(11); // DMX-512 defines minimum mab as 12us, we get 2us as the TX enables.
	UCSR0B = 0b00101000; // re-enable TXD0
}

// Initalizes start address and number of channels
void DMXinit(uint16_t address, uint16_t number) {

  DDRD = 0b00000010; /* set PD2 to output */
  PORTD = 0b11111100; // Set low, so we get a low value when TX is disabled. Magic.


	cli();
	UBRR0L = (F_CPU / 0x3d0900) - 1; 
	UBRR0H = 0;
	UCSR0C = 0b00001110;
	UCSR0B = 0b10101000; // enabled value, disabled value is UCSR0B = 0b00000000;
	UCSR0A = 0x00;
	
	sei();
	
	start_address = address;
	if(number <= CHANNELS)
	{
		num_channels = number;
	}
	else
	{
		num_channels = CHANNELS;
	}
      
        
}

// Interrupt handler for each sent data packet.    
ISR(USART_UDRE_vect)
{
	if(need_break)
	{
               
		dmx_break();
		dmx_mab();
		need_break = 0;
               
	}

	if(channel == 0 && start_packet == 0)
	{
		UDR0 = 0x00;
		start_packet = 1;
		return;
	}
	
	channel++;
	if(channel > 512)
	{
		channel = 0;
		start_packet = 0;
		need_break = 1;
		return;
	}
	
	if(channel < start_address || channel > start_address+num_channels)
	{
		UDR0 = 0x00;
	}
	else
	{
		UDR0 = universe[channel - start_address];
	}
	
	return;
}

void setup()
{
  DMXinit(1,512);
  
  pinMode(13,OUTPUT);
  
  pinMode(2,INPUT); // Configure as input (default anyway)
  digitalWrite(2,HIGH); // Enable pull-up resistor
  // Pin will read as high if nothing is connected.
   
  pinMode(3,INPUT);
  digitalWrite(3,HIGH);
  
  pinMode(4,INPUT);
  digitalWrite(4,HIGH);
  
  pinMode(5,INPUT);
  digitalWrite(5,HIGH);
  
  pinMode(6,INPUT);
  digitalWrite(6,HIGH);

  pinMode(7,INPUT);
  digitalWrite(7,HIGH);

  
}


void loop()
{

  
  // Digital pins read HIGH if nothing is connected
  // Reads LOW if connected to GND (Triggered)
  // There's no real overhead to updating the DMX address
  // This just updates the data in the buffer
  // DMX output is automatically updated on a 44Hz cycle
  // Therefore shouldn't need to debounce
  if(digitalRead(2) == LOW) // Is triggered?
  {
     DMXput(D1_ADDR,D1_CLOSED);
  } 
  else // Not triggered
  {
     DMXput(D1_ADDR,D1_OPEN);
  }
  
  if(digitalRead(3) == LOW) // Is triggered?
  {
     DMXput(D2_ADDR,D2_CLOSED);
  } 
  else // Not triggered
  {
     DMXput(D2_ADDR,D2_OPEN);
  }

  if(digitalRead(4) == LOW) // Is triggered?
  {
     DMXput(D3_ADDR,D3_CLOSED);
  } 
  else // Not triggered
  {
     DMXput(D3_ADDR,D3_OPEN);
  }

  if(digitalRead(5) == LOW) // Is triggered?
  {
     DMXput(D4_ADDR,D4_CLOSED);
  } 
  else // Not triggered
  {
     DMXput(D4_ADDR,D4_OPEN);
  }

  if(digitalRead(6) == LOW) // Is triggered?
  {
     DMXput(D5_ADDR,D5_CLOSED);
  } 
  else // Not triggered
  {
     DMXput(D5_ADDR,D5_OPEN);
  }

  if(digitalRead(7) == LOW) // Is triggered?
  {
     DMXput(D6_ADDR,D6_CLOSED);
  } 
  else // Not triggered
  {
     DMXput(D6_ADDR,D6_OPEN);
  }
}



